import os
import re
import shutil
import subprocess
import yaml
import stat
import tempfile
import textwrap
import logging
from urlparse import urlparse
from functools import partial
from charmhelpers.core import host
from charmhelpers.core import hookenv
from charmhelpers.core import services
from charmhelpers import fetch
from cloudfoundry import contexts
from cloudfoundry import templating
from cloudfoundry import utils
from .path import path
from raindance.package import PackageArchive

logger = logging.getLogger(__name__)

TEMPLATES_BASE_DIR = path('/var/vcap/jobs')
ARTIFACTS_DIR = path('/srv/artifacts')
SOFTWARE = 'cf'
ARCH = 'amd64'  # FIXME: determine from system


@hookenv.cached
def release_version(contexts=contexts):
    unit = contexts.OrchestratorRelation()['orchestrator'][0]
    return unit['cf_version']


@hookenv.cached
def job_path():
    return '/'.join(['cf-%s' % release_version(), ARCH])


def install_base_dependencies():
    fetch.apt_install(packages=fetch.filter_installed_packages([
        'ruby', 'monit', 'runit', 'zip', 'unzip', 'nfs-common']))
    gem_file = os.path.join(hookenv.charm_dir(),
                            'files/bosh-template-1.2611.0.pre.gem')
    host.adduser('vcap')
    enable_monit_http_interface()
    subprocess.check_call(['gem', 'install', '--no-ri', '--no-rdoc', gem_file])


def install(service_def):
    for job in service_def['jobs']:
        tasks = job.get('install', [])
        for task in tasks:
            if not isinstance(task, list):
                task = [task]
            for t in task:
                t()


def enable_monit_http_interface():
    enable_http = path('/etc/monit/conf.d/enable_http')

    if enable_http.exists():
        logger.warn("monit http already enabled")
        return

    enable_http.write_text(textwrap.dedent("""
        set httpd port 2812 and
           use address localhost
           allow localhost
        """))

    monit.svc_force_reload()


def install_orchestrator_key(job_name):
    rel = contexts.OrchestratorRelation()
    pub_key = rel[rel.name][0]['ssh_key']
    authorized_keys = path('/root/.ssh/authorized_keys')
    authorized_keys.write_text('\n{}'.format(pub_key), append=True)


def mount_nfs(job_name):
    orch = contexts.OrchestratorRelation()['orchestrator'][0]
    url = urlparse(orch['artifacts_url'])
    if url.scheme != 'nfs':
        return
    output = subprocess.check_output(['mount'])
    if ARTIFACTS_DIR in output:
        return
    ARTIFACTS_DIR.makedirs_p()
    subprocess.check_call([
        'mount', '-t', 'nfs', '{}:{}'.format(url.netloc, url.path), ARTIFACTS_DIR,
    ])


def fetch_job_artifacts(job_name):
    orch = contexts.OrchestratorRelation()['orchestrator'][0]
    url = orch['artifacts_url']
    version = orch['cf_version']
    if urlparse(url).scheme == 'nfs':
        return
    pa = PackageArchive(url)
    mirror = pa.build_mirror_section(ARTIFACTS_DIR, SOFTWARE, [(version, ARCH)], [job_name])
    for filename in mirror:
        pass  # just need to iterate to force the (lazy) download


def install_job_packages(base_dir, releases_dir, job_name):
    orch = contexts.OrchestratorRelation()['orchestrator'][0]
    url = orch['artifacts_url']
    version = orch['cf_version']
    pa = PackageArchive(url)
    job = pa.setup_job(ARTIFACTS_DIR, releases_dir / version,
                       SOFTWARE, version, ARCH, job_name)
    for job_file in job:
        link_source = releases_dir / version / job_file
        link_file = base_dir / job_file
        link_file.remove_p()
        link_file.makdirs_p()
        link_source.symlink(link_file)


def set_script_permissions(job_name, tmplt_base_dir=TEMPLATES_BASE_DIR):
    jobbin = tmplt_base_dir / job_name / 'bin'
    for script in jobbin.files():
        curr_mode = script.stat().st_mode
        script.chmod(curr_mode | stat.S_IEXEC)


@hookenv.cached
def load_spec(job_name):
    """
    Reads and parses the spec file for the given job name from the jobs folder.
    """
    job_extract = path(hookenv.charm_dir()) / 'jobs' / release_version() / job_name
    job_spec = job_extract / 'spec'
    with open(job_spec) as fp:
        return yaml.safe_load(fp)


def _enable_swapaccounting(s):
    output = []
    for line in s.split('\n'):
        match = re.match('^kernel\s+/boot', line)
        if match and 'swapacount' not in line:
            line = line + ' cgroup_enable=memory swapaccount=1'
        output.append(line)
    return '\n'.join(output)


def _reboot():
    subprocess.call(['shutdown', '-r', '+1'])


def in_container():
    return subprocess.call(['running-in-container']) == 0


def enable_swapaccounting(job_name):
    """To support swap accounting we must directly edit the grub
    file for the kernel cmdline. If we are already running a kerenl
    with this enabled its a no-op. Otherwise a reboot is required.
    """
    if in_container():
        return
    cmdline = open('/proc/cmdline').read()
    if "swapaccount=1" in cmdline:
        return
    grub = '/boot/grub/menu.lst'
    if not os.path.exists(grub):
        return
    shutil.copy(grub, grub + '.bak')
    source = open(grub).read()
    output = _enable_swapaccounting(source)
    with open(grub, 'w') as fp:
        fp.write(output)
    subprocess.call('update-grub')
    _reboot()


def patch_dea(job_name):
    DEA_PATCH = """
--- container.rb   2014-08-01 15:49:04.472289999 +0000
+++ container.rb 2014-07-31 23:31:30.776289999 +0000
@@ -118,7 +118,7 @@
       new_container_with_bind_mounts(params[:bind_mounts])
       limit_cpu(params[:limit_cpu])
       limit_disk(byte: params[:byte], inode: params[:inode])
-      limit_memory(params[:limit_memory])
+      #limit_memory(params[:limit_memory])
       setup_network if params[:setup_network]
     end
   end
    """
    version = release_version()
    with host.chdir('/var/vcap/releases/{}/packages/dea_next/lib/container'.format(version)):
        current = open('container.rb').read()
        if "#limit_memory(params" in current:
            # Already applied
            return
        fd, fn = tempfile.mkstemp()
        os.close(fd)
        with open(fn, 'w') as fp:
            print >>fp, DEA_PATCH
        patch = open(fn)
        subprocess.check_call(['patch', '-s', '-F4'], stdin=patch)
        os.unlink(fn)


class JobTemplates(services.ManagerCallback):
    template_base_dir = TEMPLATES_BASE_DIR

    def __init__(self, mapping):
        self.mapping = mapping

    def __call__(self, manager, job_name, event_name):
        """
        Uses the job spec to render the job's templates.
        """
        version = contexts.\
            OrchestratorRelation()['orchestrator'][0]['cf_version']
        charmdir = path(hookenv.charm_dir())
        versioned_src_dir = charmdir / 'jobs' / version / job_name
        dst_dir = self.template_base_dir / job_name
        versioned_dst_dir = self.template_base_dir / version / job_name
        templates_dir = versioned_src_dir / 'templates'
        spec = load_spec(job_name)
        callbacks = []

        for src, dst in spec.get('templates', {}).iteritems():
            versioned_dst = versioned_dst_dir / dst
            callbacks.append(templating.RubyTemplateCallback(
                src, versioned_dst, self.mapping, spec,
                templates_dir=templates_dir))

        versioned_monit_dst = versioned_dst_dir / ('monit/%s.cfg' % job_name)
        callbacks.append(templating.RubyTemplateCallback(
            'monit', versioned_monit_dst, self.mapping, spec,
            templates_dir=versioned_src_dir))

        for callback in callbacks:
            if isinstance(callback, services.ManagerCallback):
                callback(manager, job_name, event_name)
            else:
                callback(job_name)

        if dst_dir.exists():
            dst_dir.unlink()

        os.symlink(versioned_dst_dir, dst_dir)
        monit_dst = path('/etc/monit/conf.d/{}'.format(job_name))

        if monit_dst.exists():
            monit_dst.unlink()

        os.symlink(versioned_monit_dst, monit_dst)


job_templates = JobTemplates


class Monit(object):
    svc_cmd = ['service', 'monit']

    def __init__(self):
        self.name = 'monit'

    def proc(self, cmd, raise_on_err=False):
        try:
            subprocess.check_call(cmd, subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            logger.error('%s: %s', ' '.join(cmd), e.output)
            if raise_on_err:
                raise

    def get_pid(self):
        try:
            output = subprocess.check_output(['monit'])
            return int(output.split(' ')[-2])
        except (IndexError, ValueError, subprocess.CalledProcessError) as e:
            raise ValueError('Unable to parse monit pid: {}'.format(str(e)))

    def svc_restart(self, *args):
        cmd = self.svc_cmd + ['start']
        self.proc(cmd, raise_on_err=True)

    def svc_force_reload(self, *args):
        pid = self.get_pid()
        cmd = self.svc_cmd + ['force-reload']
        self.proc(cmd, raise_on_err=True)
        utils.wait_for(10, 1,
                       partial(utils.process_stopped, pid),
                       utils.monit_available)

    def reload(self, jobname):
        cmd = ['monit', 'reload']
        self.proc(cmd)

    def start(self, jobname):
        cmd = ['monit', 'restart', 'all']
        self.proc(cmd, raise_on_err=True)

    def stop(self, jobname):
        cmd = ['monit', 'stop', 'all']
        self.proc(cmd)

    def summary(self):
        try:
            output = subprocess.check_output(['monit', 'summary'])
            lines = output.split('\n')
            if len(lines) < 2:
                return None
            results = {}
            for line in lines[2:]:
                if line:
                    proc_type, name, status = line.split(None, 2)
                    results[name.strip("'")] = status
            return results
        except subprocess.CalledProcessError:
            return None


monit = Monit()
